/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Arbres binaires

#include "prelude.h"
#include <limits.h>

#include "bintree.h"

bintree *bintree_create(bintree *l, int v, bintree *r) {
  bintree *t = malloc(sizeof(struct Node));
  t->left = l;
  t->value = v;
  t->right = r;
  return t;
}

int bintree_size(bintree *t) {
  if (t == NULL) return 0;
  return 1 + bintree_size(t->left) + bintree_size(t->right);
}

// un arbre parfait de hauteur h
bintree *bintree_perfect(int h) {
  if (h == -1) return NULL;
  return bintree_create(bintree_perfect(h-1), h, bintree_perfect(h-1));
}

bintree *bintree_left(int h) {
  if (h == -1) return NULL;
  return bintree_create(bintree_left(h-1), h, NULL);
}

int max(int x, int y) {
  return x > y ? x : y;
}

// et surtout pas
// #define max(a,b) (((a) > (b)) ? (a) : (b))

int bintree_height(bintree *t) {
  if (t == NULL) return -1;
  return 1 + max(bintree_height(t->left), bintree_height(t->right));
}

void bintree_preorder(bintree *t) {
  if (t == NULL) return;
  putchar('(');
  putchar(t->value);
  bintree_preorder(t->left);
  bintree_preorder(t->right);
  putchar(')');
}
void bintree_inorder(bintree *t) {
  if (t == NULL) return;
  putchar('(');
  bintree_inorder(t->left);
  putchar(t->value);
  bintree_inorder(t->right);
  putchar(')');
}
void bintree_postorder(bintree *t) {
  if (t == NULL) return;
  putchar('(');
  bintree_postorder(t->left);
  bintree_postorder(t->right);
  putchar(t->value);
  putchar(')');
}

// exercice
bintree *bintree_random(int n) {
  if (n == 0) return NULL;
  int k = rand() % n;
  return bintree_create(bintree_random(k),
                        'a' + rand() % 26,
                        bintree_random(n - 1 - k));
}

// sérialisation depuis/vers un fichier

// comme inorder, mais dans un fichier
void bintree_write(bintree *t, FILE *f) {
  if (t == NULL) {
    fputc('#', f);
    return;
  }
  fputc('(', f);
  bintree_write(t->left, f);
  fputc(t->value, f);
  bintree_write(t->right, f);
  fputc(')', f);
}

// lit un arbre
bintree *bintree_read_aux(FILE *f) {
  bintree *l, *r;
  int x;
  int c = fgetc(f);
  if (c == '#') {
    return NULL;
  }
  assert(c == '(');
  l = bintree_read_aux(f);
  x = fgetc(f);
  r = bintree_read_aux(f);
  assert(fgetc(f) == ')');
  return bintree_create(l, x, r);
}
bintree *bintree_read(FILE *f) {
  bintree *t = bintree_read_aux(f);
  assert(fgetc(f) == EOF);
  return t;
}

// attention : suppose qu'il n'y a pas de partage de sous-arbre
void bintree_delete(bintree *t) {
  if (t == NULL) return;
  bintree_delete(t->left);
  bintree_delete(t->right);
  free(t);
}

bool bintree_isbst_aux(int lo, bintree *t, int hi) {
  if (t == NULL) return true;
  if (t->value < lo || t->value > hi) return false;
  return bintree_isbst_aux(lo, t->left, t->value)
      && bintree_isbst_aux(t->value, t->right, hi);
}
bool bintree_isbst(bintree *t) {
  return bintree_isbst_aux(INT_MIN, t, INT_MAX);
}

